package io.tswf.minecraft.blockventure.core.world.biome

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.transforms.validation.checkedmapconstructor.CheckedMapConstructor
import io.tswf.minecraft.blockventure.core.exception.BiomeEnrichException
import io.tswf.minecraft.blockventure.core.transmutation.view.ViewBiomeEffects
import io.tswf.minecraft.blockventure.core.transmutation.view.ViewGenerationSettings
import io.tswf.minecraft.blockventure.core.transmutation.view.ViewSpawnSettings
import net.minecraft.entity.EntityType
import net.minecraft.entity.SpawnGroup
import net.minecraft.registry.entry.RegistryEntry
import net.minecraft.registry.entry.RegistryEntryList
import net.minecraft.util.collection.Pool
import net.minecraft.world.biome.Biome
import net.minecraft.world.biome.SpawnSettings
import net.minecraft.world.gen.YOffset
import net.minecraft.world.gen.feature.ConfiguredFeature
import net.minecraft.world.gen.feature.PlacedFeature

import net.minecraft.world.gen.GenerationStep.Feature as FeatureGenerationStep
import net.minecraft.world.gen.placementmodifier.CountPlacementModifier
import net.minecraft.world.gen.placementmodifier.HeightRangePlacementModifier
import net.minecraft.world.gen.placementmodifier.PlacementModifier
import net.minecraft.world.gen.placementmodifier.SquarePlacementModifier

@CompileStatic
class BiomeEnricher {

    @Lazy
    static List<IBiomeEnrichProvider> biomeEnrichProviders = ServiceLoader.load(IBiomeEnrichProvider)
        .toList()

    static List<FeatureGenerationStep> FEATURE_GENERATION_STEPS = FeatureGenerationStep.values()
            .toList()
            .sortBy { it.ordinal() }

    static List<SpawnGroup> SPAWN_GROUPS = SpawnGroup.values()
        .toList()

    static boolean enrichBiome(Biome biome) {
        def enriched = false
        for (def biomeEnrichProvider : biomeEnrichProviders) {
            try {
                def enrichContext = new BiomeEnrichContext(biome: biome)
                biomeEnrichProvider.enrichBiome(enrichContext)

                def biomeGenSettings = biome.generationSettings
                def viewBiomeGenSettings = (ViewGenerationSettings) (Object) biome
                def viewBiomeEffects = (ViewBiomeEffects) (Object) biome

                def biomeSpawnSettings = biome.spawnSettings
                def viewBiomeSpawnSettings = (ViewSpawnSettings) biomeSpawnSettings

                enrichContext.fogColor?.let {
                    viewBiomeEffects.forColor = it
                    enriched = true
                }

                enrichContext.waterColor?.let {
                    viewBiomeEffects.waterColor = it
                    enriched = true
                }

                enrichContext.waterFogColor?.let {
                    viewBiomeEffects.waterFogColor = it
                    enriched = true
                }

                enrichContext.skyColor?.let {
                    viewBiomeEffects.skyColor = it
                    enriched = true
                }

                enrichContext.foliageColor?.let {
                    viewBiomeEffects.foliageColor = Optional.of(it)
                    enriched = true
                }

                enrichContext.grassColor?.let {
                    viewBiomeEffects.grassColor = Optional.of(it)
                    enriched = true
                }

                enrichContext.newFeatures?.let { newFeatures ->
                    if (newFeatures.isEmpty()) return

                    // Собираем все фичи в мапу читаемого вида. Фичи хранятся в списке списков, где первый индекс - ordinal степа, на котором они применяются. Ну и соответственно список значений.
                    def featuresBySteps = new HashMap<FeatureGenerationStep, List<RegistryEntry<PlacedFeature>>>()
                            .withDefault { (List<RegistryEntry<PlacedFeature>>) new ArrayList<RegistryEntry<PlacedFeature>>() }

                    biomeGenSettings.features.eachWithIndex{ feature, ordinal ->
                        featuresBySteps[FEATURE_GENERATION_STEPS[ordinal]]
                            .addAll(feature.stream().toList())
                    }

                    // Добавляем в список фич новые фичи
                    newFeatures.each {
                        featuresBySteps[it].addAll(it.value)
                    }


                    // Превращаем нормальную карту фич в список в исходном формате
                    def newFeaturesList = new ArrayList<RegistryEntryList<PlacedFeature>>()

                    FEATURE_GENERATION_STEPS.each { featureStep ->
                        def featuresInStep = featuresBySteps[featureStep]
                        newFeaturesList << RegistryEntryList.of(featuresInStep)
                    }

                    // Задаем новые фичи
                    viewBiomeGenSettings.features = newFeaturesList
                    enriched = true
                }

                enrichContext.newSpawnEntries?.let { newSpawnEntries ->
                    def updatedSpawners = new HashMap<SpawnGroup, Pool<SpawnSettings.SpawnEntry>>()

                    newSpawnEntries.each { spawnEntryEntry ->
                        def spawnEntriesByGroup = biomeSpawnSettings.getSpawnEntries(spawnEntryEntry.key).entries
                        updatedSpawners[spawnEntryEntry.key] = Pool.of(spawnEntryEntry.value + spawnEntriesByGroup)
                    }

                    viewBiomeSpawnSettings.spawners = updatedSpawners
                    enriched = true
                }

                enrichContext.blockEntityTypes?.let { blockEntityType ->
                    def updatedSpawners = new HashMap<SpawnGroup, Pool<SpawnSettings.SpawnEntry>>()

                    for (def spawnGroup : SPAWN_GROUPS) {
                        def spawnEntriesByGroup = biomeSpawnSettings.getSpawnEntries(spawnGroup).entries
                        updatedSpawners[spawnGroup] = Pool.of(
                                spawnEntriesByGroup.filter { it.type !in blockEntityType }
                        )
                    }

                    viewBiomeSpawnSettings.spawners = updatedSpawners
                    enriched = true
                }

                enrichContext.creatureSpawnProbability?.let {
                    viewBiomeSpawnSettings.creatureSpawnProbability = it
                    enriched = true
                }

            } catch (Throwable e) {
                throw new BiomeEnrichException(e, biome, biomeEnrichProvider)
            }
        }

        return enriched
    }
}

interface IBiomeEnrichProvider {
    void enrichBiome(BiomeEnrichContext context)
}

@CheckedMapConstructor
@CompileStatic
class BiomeEnrichContext {
    Biome biome

    // Новые фичи для генерации
    Map<FeatureGenerationStep, List<RegistryEntry<PlacedFeature>>> newFeatures = new HashMap<FeatureGenerationStep, List<RegistryEntry<PlacedFeature>>>()
            .withDefault { (List<RegistryEntry<PlacedFeature>>) new ArrayList<RegistryEntry<PlacedFeature>>() }

    // BiomeEffects
    Integer fogColor = null
    Integer waterColor = null
    Integer waterFogColor = null
    Integer skyColor = null
    Integer foliageColor = null
    Integer grassColor = null

    // Spawn
    Map<SpawnGroup, List<SpawnSettings.SpawnEntry>> newSpawnEntries = new HashMap<>()
    List<EntityType> blockEntityTypes = new ArrayList<>()

    Float creatureSpawnProbability = null

    void addFeature(FeatureGenerationStep step, RegistryEntry<PlacedFeature> feature) {
        newFeatures[step] << feature
    }

    void addFeature(FeatureGenerationStep step, PlacedFeature feature) {
        addFeature(step, new RegistryEntry.Direct<>(feature))
    }

    void addFeature(FeatureGenerationStep step, ConfiguredFeature feature, int count = 1) {
        addFeature(step, new PlacedFeature(new RegistryEntry.Direct<>(feature), [CountPlacementModifier.of(count), SquarePlacementModifier.of() ] as List<PlacementModifier>))
    }

    void addOreFeature(FeatureGenerationStep step, ConfiguredFeature feature, int count = 1, int minHeight = 0, int maxHeight = 255) {
        addFeature(step, new PlacedFeature(new RegistryEntry.Direct<>(feature), [CountPlacementModifier.of(count), SquarePlacementModifier.of(), HeightRangePlacementModifier.trapezoid(YOffset.fixed(minHeight), YOffset.fixed(maxHeight)) ] as List<PlacementModifier> ))
    }
}